#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import unittest

import numpy

from aghast import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_binning_IntegerBinning(self):
        h = Histogram([Axis(IntegerBinning(10, 20))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(11))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20"]
        assert h.axis[0].binning.toRegularBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]
        assert h.axis[0].binning.toSparseRegularBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]

        h = Histogram([Axis(IntegerBinning(10, 20, loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(13))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["(-inf, 9]", "[21, +inf)", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20"]
        assert h.axis[0].binning.toRegularBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[20.5, +inf]", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[20.5, +inf]", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[20.5, +inf]", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)"]

        h = Histogram([Axis(IntegerBinning(10, 20, loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(13))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["(-inf, 9]", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "[21, +inf)"]
        assert h.axis[0].binning.toRegularBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, 9.5)", "[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]"]

        h = Histogram([Axis(IntegerBinning(10, 20, loc_underflow=RealOverflow.above2, loc_overflow=RealOverflow.above1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(13))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "[21, +inf)", "(-inf, 9]"]
        assert h.axis[0].binning.toRegularBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]", "[-inf, 9.5)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]", "[-inf, 9.5)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[9.5, 10.5)", "[10.5, 11.5)", "[11.5, 12.5)", "[12.5, 13.5)", "[13.5, 14.5)", "[14.5, 15.5)", "[15.5, 16.5)", "[16.5, 17.5)", "[17.5, 18.5)", "[18.5, 19.5)", "[19.5, 20.5)", "[20.5, +inf]", "[-inf, 9.5)"]

    def test_binning_RegularBinning(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(0.1, 10.1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)", "[9.1, 10.1)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)", "[9.1, 10.1)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)", "[9.1, 10.1)"]
        assert h.axis[0].binning.toSparseRegularBinning().toCategoryBinning().categories == ["[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)", "[9.1, 10.1)"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-0.9, 9.1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-0.9, 0.1)", "[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)"]
        assert h.axis[0].binning.toSparseRegularBinning().toCategoryBinning().categories == ["[-0.9, 0.1)", "[0.1, 1.1)", "[1.1, 2.1)", "[2.1, 3.1)", "[3.1, 4.1)", "[4.1, 5.1)", "[5.1, 6.1)", "[6.1, 7.1)", "[7.1, 8.1)", "[8.1, 9.1)"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-100, 100)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]
        assert h.axis[0].binning.toSparseRegularBinning().toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-100, 100), overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(12))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, -100)", "[100, +inf]", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-inf, -100)", "[100, +inf]", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, -100)", "[100, +inf]", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-100, 100), overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(12))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-100, 100), overflow=RealOverflow(loc_underflow=RealOverflow.above2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(12))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]", "[-inf, -100)"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]", "[-inf, -100)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf]", "[-inf, -100)"]

        h = Histogram([Axis(RegularBinning(10, RealInterval(-100, 100), overflow=RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.above1, loc_nanflow=RealOverflow.above2, minf_mapping=RealOverflow.in_nanflow, pinf_mapping=RealOverflow.in_nanflow)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(13))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["(-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf)", "{-inf, +inf, nan}"]
        assert h.axis[0].binning.toEdgesBinning().toCategoryBinning().categories == ["(-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf)", "{-inf, +inf, nan}"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["(-inf, -100)", "[-100, -80)", "[-80, -60)", "[-60, -40)", "[-40, -20)", "[-20, 0)", "[0, 20)", "[20, 40)", "[40, 60)", "[60, 80)", "[80, 100)", "[100, +inf)", "{-inf, +inf, nan}"]

    def test_binning_EdgesBinning(self):
        h = Histogram([Axis(EdgesBinning([3, 4.5, 10, 20]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)"]

        h = Histogram([Axis(EdgesBinning([3, 4.5, 10, 20], overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, 3)", "[20, +inf]", "[3, 4.5)", "[4.5, 10)", "[10, 20)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, 3)", "[20, +inf]", "[3, 4.5)", "[4.5, 10)", "[10, 20)"]

        h = Histogram([Axis(EdgesBinning([3, 4.5, 10, 20], overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]"]

        h = Histogram([Axis(EdgesBinning([3, 4.5, 10, 20], overflow=RealOverflow(loc_underflow=RealOverflow.above2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]", "[-inf, 3)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]", "[-inf, 3)"]

        h = Histogram([Axis(EdgesBinning([3, 4.5, 10, 20], overflow=RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.above1, loc_nanflow=RealOverflow.above2, minf_mapping=RealOverflow.in_nanflow, pinf_mapping=RealOverflow.in_nanflow)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(6))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["(-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf)", "{-inf, +inf, nan}"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["(-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf)", "{-inf, +inf, nan}"]

    def test_binning_IrregularBinning(self):
        h = Histogram([Axis(IrregularBinning([RealInterval(3, 4.5), RealInterval(4.5, 10), RealInterval(10, 20)]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(3))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)"]

        h = Histogram([Axis(IrregularBinning([RealInterval(3, 4.5), RealInterval(4.5, 10), RealInterval(10, 20)], overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, 3)", "[20, +inf]", "[3, 4.5)", "[4.5, 10)", "[10, 20)"]

        h = Histogram([Axis(IrregularBinning([RealInterval(3, 4.5), RealInterval(4.5, 10), RealInterval(10, 20)], overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]"]

        h = Histogram([Axis(IrregularBinning([RealInterval(3, 4.5), RealInterval(4.5, 10), RealInterval(10, 20)], overflow=RealOverflow(loc_underflow=RealOverflow.above2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf]", "[-inf, 3)"]

        h = Histogram([Axis(IrregularBinning([RealInterval(3, 4.5), RealInterval(4.5, 10), RealInterval(10, 20)], overflow=RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.above1, loc_nanflow=RealOverflow.above2, minf_mapping=RealOverflow.in_nanflow, pinf_mapping=RealOverflow.in_nanflow)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(6))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["(-inf, 3)", "[3, 4.5)", "[4.5, 10)", "[10, 20)", "[20, +inf)", "{-inf, +inf, nan}"]

    def test_binning_SparseRegularBinning(self):
        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(5))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-29.9, -19.9)", "[60.1, 70.1)", "[100.1, 110.1)", "[110.1, 120.1)", "[120.1, 130.1)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["[-29.9, -19.9)", "[60.1, 70.1)", "[100.1, 110.1)", "[110.1, 120.1)", "[120.1, 130.1)"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(7))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{-inf}", "{+inf}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(7))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{-inf}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)", "{+inf}"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_underflow=RealOverflow.above2, loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(7))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)", "{+inf}", "{-inf}"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_underflow=RealOverflow.below2, loc_overflow=RealOverflow.above1, loc_nanflow=RealOverflow.above2)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(8))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{-inf}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)", "{+inf}", "{nan}"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_nanflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(6))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{nan}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["{nan}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_nanflow=RealOverflow.below1, minf_mapping=RealOverflow.in_nanflow, pinf_mapping=RealOverflow.in_nanflow)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(6))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{-inf, +inf, nan}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["{-inf, +inf, nan}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]

        h = Histogram([Axis(SparseRegularBinning([-3, 6, 10, 11, 12], 10, 0.0, overflow=RealOverflow(loc_nanflow=RealOverflow.below1, minf_mapping=RealOverflow.in_nanflow, pinf_mapping=RealOverflow.in_nanflow, nan_mapping=RealOverflow.missing)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(6))))
        assert h.axis[0].binning.toCategoryBinning().categories == ["{-inf, +inf}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]
        assert h.axis[0].binning.toIrregularBinning().toCategoryBinning().categories == ["{-inf, +inf}", "[-30, -20)", "[60, 70)", "[100, 110)", "[110, 120)", "[120, 130)"]



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Column(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsColumn(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Column()
        x.Init(buf, n + offset)
        return x

    # Column
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Column
    def Identifier(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Column
    def Dtype(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Column
    def Endianness(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Column
    def Filters(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Column
    def FiltersAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Column
    def FiltersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Column
    def PostfilterSlice(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .Slice import Slice
            obj = Slice()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Column
    def Title(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Column
    def Metadata(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Metadata import Metadata
            obj = Metadata()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Column
    def Decoration(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Decoration import Decoration
            obj = Decoration()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def ColumnStart(builder): builder.StartObject(8)
def ColumnAddIdentifier(builder, identifier): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(identifier), 0)
def ColumnAddDtype(builder, dtype): builder.PrependInt8Slot(1, dtype, 0)
def ColumnAddEndianness(builder, endianness): builder.PrependInt8Slot(2, endianness, 0)
def ColumnAddFilters(builder, filters): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(filters), 0)
def ColumnStartFiltersVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ColumnAddPostfilterSlice(builder, postfilterSlice): builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(postfilterSlice), 0)
def ColumnAddTitle(builder, title): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(title), 0)
def ColumnAddMetadata(builder, metadata): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0)
def ColumnAddDecoration(builder, decoration): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(decoration), 0)
def ColumnEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import pickle
import unittest

import numpy

from aghast import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_pickle(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))))
        assert h == pickle.loads(pickle.dumps(h))

    def test_serialization_Metadata(self):
        h = Collection({}, metadata=Metadata("""{"one": 1, "two": 2}""", language=Metadata.json))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Collection({}, metadata=Metadata("""{"one": 1, "two": 2}"""))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Decoration(self):
        h = Collection({}, decoration=Decoration("""points { color: red }""", language=Decoration.css))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Collection({}, decoration=Decoration("""points { color: red }"""))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_RawInlineBuffer(self):
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(numpy.zeros(1, dtype=numpy.int32)))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_RawExternalBuffer(self):
        buf = numpy.zeros(1, dtype=numpy.int32)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawExternalBuffer(buf.ctypes.data, buf.nbytes))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        buf = numpy.array([3.14], dtype=numpy.float64)
        h = Ntuple([Column("one", Column.float64)], [NtupleInstance([Chunk([ColumnChunk([Page(RawExternalBuffer(buf.ctypes.data, buf.nbytes))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_InterpretedInlineBuffer(self):
        h = BinnedEvaluatedFunction([Axis()], InterpretedInlineBuffer(numpy.zeros(1, dtype=numpy.int32), dtype=InterpretedInlineBuffer.int32))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis()], InterpretedInlineBuffer(b"\x07\x00\x00\x00", dtype=InterpretedInlineBuffer.int32))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_InterpretedExternalBuffer(self):
        buf = numpy.zeros(1, dtype=numpy.float64)
        h = BinnedEvaluatedFunction([Axis()], InterpretedExternalBuffer(buf.ctypes.data, buf.nbytes, dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        buf = numpy.array([3.14], dtype=numpy.float64)
        h = BinnedEvaluatedFunction([Axis()], InterpretedExternalBuffer(buf.ctypes.data, buf.nbytes, dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_IntegerBinning(self):
        h = Histogram([Axis(IntegerBinning(10, 20))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(11))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(IntegerBinning(10, 20, loc_underflow=IntegerBinning.above1))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(12))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(IntegerBinning(10, 20, loc_underflow=IntegerBinning.above1, loc_overflow=IntegerBinning.above2))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(13))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_RealInterval(self):
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5)))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_RealOverflow(self):
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow()))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.nonexistent, loc_nanflow=RealOverflow.nonexistent)))], InterpretedInlineBuffer(numpy.zeros(11), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.below2, loc_nanflow=RealOverflow.nonexistent)))], InterpretedInlineBuffer(numpy.zeros(12), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(loc_underflow=RealOverflow.below1, loc_overflow=RealOverflow.below2, loc_nanflow=RealOverflow.below3)))], InterpretedInlineBuffer(numpy.zeros(13), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(minf_mapping=RealOverflow.in_nanflow)))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(pinf_mapping=RealOverflow.in_nanflow)))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5), RealOverflow(nan_mapping=RealOverflow.missing)))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_RegularBinning(self):
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5)))], InterpretedInlineBuffer(numpy.zeros(10), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_HexagonalBinning(self):
        h = BinnedEvaluatedFunction([Axis(HexagonalBinning(3, 5, -5, -4))], InterpretedInlineBuffer(numpy.array([[0.0] * 2] * 3), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(HexagonalBinning(3, 5, -5, -4, qoverflow=RealOverflow(loc_nanflow=RealOverflow.above1)))], InterpretedInlineBuffer(numpy.array([[0.0] * 2] * 4), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(HexagonalBinning(3, 5, -5, -4, roverflow=RealOverflow(loc_nanflow=RealOverflow.above1)))], InterpretedInlineBuffer(numpy.array([[0.0] * 3] * 3), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(HexagonalBinning(3, 5, -5, -4, qoverflow=RealOverflow(loc_nanflow=RealOverflow.above1), roverflow=RealOverflow(loc_nanflow=RealOverflow.above1)))], InterpretedInlineBuffer(numpy.array([[0.0] * 3] * 4), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_EdgesBinning(self):
        h = BinnedEvaluatedFunction([Axis(EdgesBinning([3.3], overflow=RealOverflow(loc_underflow=RealOverflow.above1, loc_overflow=RealOverflow.above2)))], InterpretedInlineBuffer(numpy.array([0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(EdgesBinning([1.1, 2.2, 3.3]))], InterpretedInlineBuffer(numpy.array([0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_IrregularBinning(self):
        h = BinnedEvaluatedFunction([Axis(IrregularBinning([RealInterval(0.5, 1.5)]))], InterpretedInlineBuffer(numpy.array([0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(IrregularBinning([RealInterval(0.5, 1.5), RealInterval(1.5, 1.5), RealInterval(0.0, 10.0)]))], InterpretedInlineBuffer(numpy.array([0.0, 0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_CategoryBinning(self):
        h = BinnedEvaluatedFunction([Axis(CategoryBinning(["one", "two", "three"]))], InterpretedInlineBuffer(numpy.array([0.0, 0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(CategoryBinning(["one", "two", "three"], loc_overflow=CategoryBinning.above1))], InterpretedInlineBuffer(numpy.array([0.0, 0.0, 0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_SparseRegularBinning(self):
        h = BinnedEvaluatedFunction([Axis(SparseRegularBinning([-5, -3, 10, 1000], 0.1))], InterpretedInlineBuffer(numpy.array([0.0, 0.0, 0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_FractionBinning(self):
        h = BinnedEvaluatedFunction([Axis(FractionBinning())], InterpretedInlineBuffer(numpy.array([0.0, 0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(FractionBinning()), Axis(RegularBinning(10, RealInterval(-5, 5)))], InterpretedInlineBuffer(numpy.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = BinnedEvaluatedFunction([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(FractionBinning())], InterpretedInlineBuffer(numpy.array([[[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]]]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_PredicateBinning(self):
        h = Histogram([Axis(PredicateBinning(["p", "q"]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0.0, 0.0]))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Assignments(self):
        h = Histogram([Axis(VariationBinning([Variation([Assignment("x", "1"), Assignment("y", "2"), Assignment("z", "3")])]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Variation(self):
        h = Histogram([Axis(VariationBinning([Variation([Assignment("x", "1")]), Variation([Assignment("x", "2")])]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0.0, 0.0]))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_VariationBinning(self):
        h = Histogram([Axis(VariationBinning([Variation([Assignment("x", "1")]), Variation([Assignment("x", "2")]), Variation([Assignment("x", "3")])]))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0.0, 0.0, 0.0]))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Axis(self):
        h = BinnedEvaluatedFunction([Axis(expression="x", title="wow")], InterpretedInlineBuffer(numpy.array([0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_UnweightedCounts(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_WeightedCounts(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], WeightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], WeightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10)), sumw2=InterpretedInlineBuffer.fromarray(numpy.arange(10)**2)))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], WeightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10)), sumw2=InterpretedInlineBuffer.fromarray(numpy.arange(10)**2), unweighted=UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10)))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_StatisticFilter(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics()])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 1, filter=StatisticFilter(excludes_nan=False))])])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Moments(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 2)])])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 0, weightpower=0), Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 0, weightpower=1)])])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Extremes(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(min=Extremes(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))))])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(min=Extremes(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))), max=Extremes(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))))])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Quantiles(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(quantiles=[Quantiles(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 0.25), Quantiles(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))), Quantiles(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 0.75)])])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(quantiles=[Quantiles(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), weightpower=0), Quantiles(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), weightpower=1)])])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Modes(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics(mode=Modes(InterpretedInlineBuffer.fromarray(numpy.array([0.0]))))])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Statistics(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics()]), Axis(RegularBinning(10, RealInterval(-5, 5)), statistics=[Statistics()])], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32), Column("two", Column.int16)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1]), ColumnChunk([Page(RawInlineBuffer(b"\x03\x00"))], [0, 1])])])], column_statistics=[Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.array([0.0])), 2)])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Covariance(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), axis_covariances=[Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1)))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(1000))), axis_covariances=[Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1))), Covariance(0, 2, InterpretedInlineBuffer.fromarray(numpy.arange(1))), Covariance(1, 2, InterpretedInlineBuffer.fromarray(numpy.arange(1)))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))), [Profile("", Statistics([Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 2)])), Profile("", Statistics([Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 2)]))], profile_covariances=[Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1)))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32), Column("two", Column.int16)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1]), ColumnChunk([Page(RawInlineBuffer(b"\x03\x00"))], [0, 1])])])], column_covariances=[Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1)))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32), Column("two", Column.int16)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1]), ColumnChunk([Page(RawInlineBuffer(b"\x03\x00"))], [0, 1])])])], column_covariances=[Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1)), weightpower=1), Covariance(0, 1, InterpretedInlineBuffer.fromarray(numpy.arange(1)), weightpower=0)])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Profile(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10))), [Profile("", Statistics([Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(10)), 2)]))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), [Profile("", Statistics([Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(100)), 1), Moments(InterpretedInlineBuffer.fromarray(numpy.zeros(100)), 2)]))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Histogram(self):
        h = Histogram([Axis()], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(1))))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Parameter(self):
        h = ParameterizedFunction("x**2", [Parameter("x", InterpretedInlineBuffer.fromarray(numpy.array([5]))), Parameter("y", InterpretedInlineBuffer.fromarray(numpy.array([6])))])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_ParameterizedFunction(self):
        h = ParameterizedFunction("x**2")
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), functions={"f": ParameterizedFunction("x**2", [Parameter("x", InterpretedInlineBuffer.fromarray(numpy.arange(100)))])})
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_EvaluatedFunction(self):
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), functions={"f": EvaluatedFunction(InterpretedInlineBuffer.fromarray(numpy.arange(100)))})
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), functions={"id": EvaluatedFunction(InterpretedInlineBuffer.fromarray(numpy.arange(100)), InterpretedInlineBuffer.fromarray(numpy.arange(100)))})
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5))), Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(100))), functions={"id": EvaluatedFunction(InterpretedInlineBuffer.fromarray(numpy.arange(100)), InterpretedInlineBuffer.fromarray(numpy.arange(100)), [Quantiles(InterpretedInlineBuffer.fromarray(numpy.zeros(100)), 0.25), Quantiles(InterpretedInlineBuffer.fromarray(numpy.zeros(100))), Quantiles(InterpretedInlineBuffer.fromarray(numpy.zeros(100)), 0.75)])})
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_BinnedEvaluatedFunction(self):
        h = BinnedEvaluatedFunction([Axis()], InterpretedInlineBuffer(numpy.array([0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        assert frombuffer(h.tobuffer()).values.array.tolist() == [0.0]
        h = BinnedEvaluatedFunction([Axis(), Axis()], InterpretedInlineBuffer(numpy.array([0.0]), dtype=InterpretedInlineBuffer.float64))
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        assert frombuffer(h.tobuffer()).values.array.tolist() == [[0.0]]

    def test_serialization_Page(self):
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([], [0])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00")), Page(RawInlineBuffer(b"\x04\x00\x00\x00\x03\x00\x00\x00"))], [0, 1, 3])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Chunk(self):
        h = Ntuple([Column("one", Column.float64)], [NtupleInstance([Chunk([ColumnChunk([], [0])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])]), Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Column(self):
        h = Ntuple([Column("one", Column.float64), Column("two", Column.int32)], [NtupleInstance([])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Ntuple([Column("one", Column.int32), Column("two", Column.int16)], [NtupleInstance([Chunk([ColumnChunk([Page(RawInlineBuffer(b"\x05\x00\x00\x00"))], [0, 1]), ColumnChunk([Page(RawInlineBuffer(b"\x03\x00"))], [0, 1])])])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_Ntuple(self):
        h = Ntuple([Column("one", Column.float64)], [NtupleInstance([])])
        assert h == frombuffer(h.tobuffer(), checkvalid=True)

    def test_serialization_collection(self):
        # h = Collection()
        # assert h == frombuffer(h.tobuffer(), checkvalid=True)
        h = Collection({"b": Collection(), "c": Collection()})
        assert h == frombuffer(h.tobuffer(), checkvalid=True)
        # h = Collection({"id": Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(30)))), "id2": Histogram([Axis(RegularBinning(100, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(300))))}, axis=[Axis(RegularBinning(3, RealInterval(-1, 1)))])
        # assert h == frombuffer(h.tobuffer(), checkvalid=True)
        # h = Collection({"b": Collection({"c": Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(60)))), "d": Histogram([Axis(RegularBinning(100, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(600))))}, axis=[Axis(FractionBinning())])}, axis=[Axis(RegularBinning(3, RealInterval(-1, 1)))])
        # assert h == frombuffer(h.tobuffer(), checkvalid=True)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

# ///////////////////////////////////////////////// objects and collections
class Collection(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCollection(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Collection()
        x.Init(buf, n + offset)
        return x

    # Collection
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Collection
    def Lookup(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Collection
    def LookupLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Collection
    def Objects(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Object import Object
            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collection
    def ObjectsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Collection
    def Axis(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Axis import Axis
            obj = Axis()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collection
    def AxisLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def CollectionStart(builder): builder.StartObject(3)
def CollectionAddLookup(builder, lookup): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(lookup), 0)
def CollectionStartLookupVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CollectionAddObjects(builder, objects): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(objects), 0)
def CollectionStartObjectsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CollectionAddAxis(builder, axis): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(axis), 0)
def CollectionStartAxisVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def CollectionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class HexagonalCoordinates(object):
    hex_offset = 0
    hex_doubled_offset = 1
    hex_cube_xy = 2
    hex_cube_yz = 3
    hex_cube_xz = 4




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Slice(object):
    __slots__ = ['_tab']

    # Slice
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Slice
    def Start(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Slice
    def Stop(self): return self._tab.Get(flatbuffers.number_types.Int64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # Slice
    def Step(self): return self._tab.Get(flatbuffers.number_types.Int32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # Slice
    def HasStart(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(20))
    # Slice
    def HasStop(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(21))
    # Slice
    def HasStep(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(22))

def CreateSlice(builder, start, stop, step, hasStart, hasStop, hasStep):
    builder.Prep(8, 24)
    builder.Pad(1)
    builder.PrependBool(hasStep)
    builder.PrependBool(hasStop)
    builder.PrependBool(hasStart)
    builder.PrependInt32(step)
    builder.PrependInt64(stop)
    builder.PrependInt64(start)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class RealOverflow(object):
    __slots__ = ['_tab']

    # RealOverflow
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RealOverflow
    def LocUnderflow(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # RealOverflow
    def LocOverflow(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(1))
    # RealOverflow
    def LocNanflow(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(2))
    # RealOverflow
    def MinfMapping(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(3))
    # RealOverflow
    def PinfMapping(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # RealOverflow
    def NanMapping(self): return self._tab.Get(flatbuffers.number_types.Int8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(5))

def CreateRealOverflow(builder, locUnderflow, locOverflow, locNanflow, minfMapping, pinfMapping, nanMapping):
    builder.Prep(1, 6)
    builder.PrependInt8(nanMapping)
    builder.PrependInt8(pinfMapping)
    builder.PrependInt8(minfMapping)
    builder.PrependInt8(locNanflow)
    builder.PrependInt8(locOverflow)
    builder.PrependInt8(locUnderflow)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

# ///////////////////////////////////////////////// descriptive statistics
class StatisticFilter(object):
    __slots__ = ['_tab']

    # StatisticFilter
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # StatisticFilter
    def Min(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # StatisticFilter
    def Max(self): return self._tab.Get(flatbuffers.number_types.Float64Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # StatisticFilter
    def ExcludesMinf(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # StatisticFilter
    def ExcludesPinf(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(17))
    # StatisticFilter
    def ExcludesNan(self): return self._tab.Get(flatbuffers.number_types.BoolFlags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(18))

def CreateStatisticFilter(builder, min, max, excludesMinf, excludesPinf, excludesNan):
    builder.Prep(8, 24)
    builder.Pad(5)
    builder.PrependBool(excludesNan)
    builder.PrependBool(excludesPinf)
    builder.PrependBool(excludesMinf)
    builder.PrependFloat64(max)
    builder.PrependFloat64(min)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Object(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsObject(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Object()
        x.Init(buf, n + offset)
        return x

    # Object
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Object
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Object
    def Data(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Object
    def Title(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Object
    def Metadata(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Metadata import Metadata
            obj = Metadata()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Object
    def Decoration(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .Decoration import Decoration
            obj = Decoration()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Object
    def Script(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def ObjectStart(builder): builder.StartObject(6)
def ObjectAddDataType(builder, dataType): builder.PrependUint8Slot(0, dataType, 0)
def ObjectAddData(builder, data): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def ObjectAddTitle(builder, title): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(title), 0)
def ObjectAddMetadata(builder, metadata): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0)
def ObjectAddDecoration(builder, decoration): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(decoration), 0)
def ObjectAddScript(builder, script): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(script), 0)
def ObjectEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class NtupleInstance(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsNtupleInstance(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NtupleInstance()
        x.Init(buf, n + offset)
        return x

    # NtupleInstance
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # NtupleInstance
    def Chunks(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Chunk import Chunk
            obj = Chunk()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # NtupleInstance
    def ChunksLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NtupleInstance
    def ChunkOffsets(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # NtupleInstance
    def ChunkOffsetsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # NtupleInstance
    def ChunkOffsetsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def NtupleInstanceStart(builder): builder.StartObject(2)
def NtupleInstanceAddChunks(builder, chunks): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(chunks), 0)
def NtupleInstanceStartChunksVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def NtupleInstanceAddChunkOffsets(builder, chunkOffsets): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(chunkOffsets), 0)
def NtupleInstanceStartChunkOffsetsVector(builder, numElems): return builder.StartVector(8, numElems, 8)
def NtupleInstanceEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class Binning(object):
    NONE = 0
    IntegerBinning = 1
    RegularBinning = 2
    HexagonalBinning = 3
    EdgesBinning = 4
    IrregularBinning = 5
    CategoryBinning = 6
    SparseRegularBinning = 7
    FractionBinning = 8
    PredicateBinning = 9
    VariationBinning = 10




#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import numpy

from aghast import *
import aghast.interface

def binning2array(binning):
    if not isinstance(binning, EdgesBinning):
        if not hasattr(binning, "toEdgesBinning"):
            raise TypeError("cannot convert {0} to a Numpy binning".format(type(binning).__name__))
        binning = binning.toEdgesBinning()
    if binning.overflow is not None and binning.overflow.loc_underflow != BinLocation.nonexistent and binning.overflow.loc_overflow != BinLocation.nonexistent:
        out = numpy.empty(len(binning.edges) + 2, dtype=binning.edges.dtype)
        out[0] = -numpy.inf
        out[-1] = numpy.inf
        out[1:-1] = binning.edges
    elif binning.overflow is not None and binning.overflow.loc_underflow != BinLocation.nonexistent:
        out = numpy.empty(len(binning.edges) + 1, dtype=binning.edges.dtype)
        out[0] = -numpy.inf
        out[1:] = binning.edges
    elif binning.overflow is not None and binning.overflow.loc_overflow != BinLocation.nonexistent:
        out = numpy.empty(len(binning.edges) + 1, dtype=binning.edges.dtype)
        out[-1] = numpy.inf
        out[:-1] = binning.edges
    else:
        out = binning.edges
    return out
        
def to_numpy(obj):
    if isinstance(obj, Histogram):
        edges = [binning2array(x.binning) for x in obj.axis]
        slices = ()
        for x in edges:
            start = -numpy.inf if x[0] == -numpy.inf else None
            stop = numpy.inf if x[-1] == numpy.inf else None
            slices = slices + (slice(start, stop),)
        counts = obj.counts[slices]
        if isinstance(counts, dict):
            counts = counts["sumw"]

        if len(edges) == 1:
            return counts, edges[0]
        elif len(edges) == 2:
            return counts, edges[0], edges[1]
        else:
            return counts, edges

    else:
        raise TypeError("cannot convert {0} to a Numpy histogram".format(type(obj).__name__))

def array2counts(array):
    if issubclass(array.dtype.type, numpy.integer) and (array >= 0).all():
        return UnweightedCounts(InterpretedInlineBuffer.fromarray(array))
    else:
        return WeightedCounts(InterpretedInlineBuffer.fromarray(array))

def array2binning(array):
    if len(array) <= 1:
        raise ValueError("binning array must have at least 2 elements: {0}".format(repr(array)))
    if not (array[1:] >= array[:-1]).all():
        raise ValueError("binning array must be monotonically increasing: {0}".format(repr(array)))
    bin_widths = array[1:] - array[:-1]
    bin_width = bin_widths.mean()
    if (numpy.absolute(bin_widths - bin_width) < 1e-10*(array[-1] - array[0])).all():
        return RegularBinning(len(array) - 1, RealInterval(array[0], array[-1]))
    else:
        return EdgesBinning(array)

def from_numpy(obj):
    if isinstance(obj, tuple) and len(obj) == 2 and isinstance(obj[0], numpy.ndarray) and isinstance(obj[1], numpy.ndarray):
        counts, edges = obj
        return Histogram([Axis(array2binning(edges))], array2counts(counts))

    elif isinstance(obj, tuple) and len(obj) == 3 and isinstance(obj[0], numpy.ndarray) and isinstance(obj[1], numpy.ndarray) and isinstance(obj[2], numpy.ndarray):
        counts, xedges, yedges = obj
        return Histogram([Axis(array2binning(xedges)), Axis(array2binning(yedges))], array2counts(counts))

    elif isinstance(obj, tuple) and len(obj) == 2 and isinstance(obj[0], numpy.ndarray) and isinstance(obj[1], list) and all(isinstance(x, numpy.ndarray) for x in obj[1]):
        counts, edges = obj
        return Histogram([Axis(array2binning(x)) for x in edges], array2counts(counts))

    else:
        raise TypeError("not a recognized Numpy histogram type")



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import sys
import unittest

import numpy

from aghast import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_getitem_twodim(self):
        a = Histogram([Axis(IntegerBinning(0, 3)), Axis(IntegerBinning(0, 2))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([[10, 100, 1000], [20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]))))
        a.checkvalid()
        assert a.axis[0].binning.toCategoryBinning().categories == ["0", "1", "2", "3"]
        assert a.axis[1].binning.toCategoryBinning().categories == ["0", "1", "2"]
        assert a.counts.counts.array.tolist() == [[10, 100, 1000], [20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]

        assert a.counts[None, None] == sum([10, 100, 1000, 20, 200, 2000, 30, 300, 3000, 40, 400, 4000])
        assert a.counts[None, :].tolist() == [100, 1000, 10000]
        assert a.counts[None].tolist() == [100, 1000, 10000]
        assert a.counts[:, None].tolist() == [1110, 2220, 3330, 4440]
        assert a.counts[None, 1] == 1000
        assert a.counts[1, None] == 2220
        assert a.counts[None, 1:].tolist() == [1000, 10000]
        assert a.counts[1:, None].tolist() == [2220, 3330, 4440]
        assert a.counts[None, [2, 1, 1, 0]].tolist() == [10000, 1000, 1000, 100]
        assert a.counts[[3, 2, 2, 0], None].tolist() == [4440, 3330, 3330, 1110]
        assert a.counts[None, [True, False, True]].tolist() == [100, 10000]
        assert a.counts[[False, True, True, False], None].tolist() == [2220, 3330]

        assert a.counts[:, :].tolist() == [[10, 100, 1000], [20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]
        assert a.counts[:].tolist() == [[10, 100, 1000], [20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]
        assert a.counts[1:, :].tolist() == [[20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]
        assert a.counts[1:].tolist() == [[20, 200, 2000], [30, 300, 3000], [40, 400, 4000]]
        assert a.counts[:, 1:].tolist() == [[100, 1000], [200, 2000], [300, 3000], [400, 4000]]
        assert a.counts[2:, 1:].tolist() == [[300, 3000], [400, 4000]]
        assert a.counts[:, 1].tolist() == [100, 200, 300, 400]
        assert a.counts[1, :].tolist() == [20, 200, 2000]
        assert a.counts[1].tolist() == [20, 200, 2000]
        assert a.counts[2:, 1].tolist() == [300, 400]
        assert a.counts[1, 2:].tolist() == [2000]
        assert a.counts[:, [2, 0]].tolist() == [[1000, 10], [2000, 20], [3000, 30], [4000, 40]]
        assert a.counts[[2, 0], :].tolist() == [[30, 300, 3000], [10, 100, 1000]]
        assert a.counts[1:, [2, 0]].tolist() == [[2000, 20], [3000, 30], [4000, 40]]
        assert a.counts[[2, 0], 1:].tolist() == [[300, 3000], [100, 1000]]
        assert a.counts[:, [True, False, True]].tolist() == [[10, 1000], [20, 2000], [30, 3000], [40, 4000]]
        assert a.counts[[False, True, True, False], :].tolist() == [[20, 200, 2000], [30, 300, 3000]]
        assert a.counts[1:, [True, False, True]].tolist() == [[20, 2000], [30, 3000], [40, 4000]]
        assert a.counts[[False, True, True, False], 1:].tolist() == [[200, 2000], [300, 3000]]

        assert a.counts[1, 2] == 2000
        assert a.counts[1, [2, 2, 0]].tolist() == [2000, 2000, 20]
        assert a.counts[[2, 2, 0], 1].tolist() == [300, 300, 100]
        assert a.counts[1, [True, False, True]].tolist() == [20, 2000]
        assert a.counts[[False, True, True, False], 1].tolist() == [200, 300]

        assert a.counts[[1, 2], [2, 0]].tolist() == [[2000, 20], [3000, 30]]
        assert a.counts[[False, True, True, False], [2, 0]].tolist() == [[2000, 20], [3000, 30]]
        assert a.counts[[False, True, True, False], [True, False, True]].tolist() == [[20, 2000], [30, 3000]]

        assert a.counts[[2, 0], [2, 2, 0]].tolist() == [[3000, 3000, 30], [1000, 1000, 10]]
        assert a.counts[[2, 0], [True, False, True]].tolist() == [[30, 3000], [10, 1000]]
        assert a.counts[[True, False, True, False], [True, False, True]].tolist() == [[10, 1000], [30, 3000]]

    def test_getitem_IntegerBinning(self):
        a = Histogram([Axis(IntegerBinning(-5, 5))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(11, dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        assert a.counts[None] == 55
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9, 10]
        assert a.counts[5] == 5
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[numpy.array([7, 4, 7, 5, -1])].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False, True]].tolist() == [0, 2, 4, 6, 8, 10]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False, True])].tolist() == [0, 2, 4, 6, 8, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5, loc_overflow=IntegerBinning.above1))], UnweightedCounts(InterpretedInlineBuffer.fromarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999])))
        assert a.axis[0].binning.toCategoryBinning().categories == ["-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5", "[6, +inf)"]
        assert a.counts.counts.array.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999]

        assert a.counts[None] == 55 + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9, 10]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 10, 999]
        assert a.counts[5] == 5
        assert a.counts[numpy.inf] == 999
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -1]].tolist() == [7, 4, 7, 999, 5, 10]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False, True]].tolist() == [0, 2, 4, 6, 8, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5, loc_overflow=IntegerBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray([999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[6, +inf)", "-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        assert a.counts[None] == 55 + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9, 10]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 10, 999]
        assert a.counts[5] == 5
        assert a.counts[numpy.inf] == 999
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -1]].tolist() == [7, 4, 7, 999, 5, 10]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False, True]].tolist() == [0, 2, 4, 6, 8, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5, loc_underflow=IntegerBinning.below2, loc_overflow=IntegerBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray([123, 999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])))
        assert a.axis[0].binning.toCategoryBinning().categories == ["(-inf, -6]", "[6, +inf)", "-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5"]
        assert a.counts.counts.array.tolist() == [123, 999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        assert a.counts[None] == 55 + 123 + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9, 10]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 10, 999]
        assert a.counts[-numpy.inf:5].tolist() == [123, 0, 1, 2, 3, 4]
        assert a.counts[-numpy.inf:numpy.inf].tolist() == [123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999]
        assert a.counts[-numpy.inf:].tolist() == [123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5] == 5
        assert a.counts[-numpy.inf] == 123
        assert a.counts[numpy.inf] == 999
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -1]].tolist() == [7, 4, 7, 999, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -numpy.inf, -1]].tolist() == [7, 4, 7, 999, 5, 123, 10]
        assert a.counts[[7, -numpy.inf, 4, 7, numpy.inf, 5, -numpy.inf, -1]].tolist() == [7, 123, 4, 7, 999, 5, 123, 10]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False, True]].tolist() == [0, 2, 4, 6, 8, 10]

        a = Histogram([Axis(IntegerBinning(-5, 5, loc_underflow=IntegerBinning.above1, loc_overflow=IntegerBinning.below1))], UnweightedCounts(InterpretedInlineBuffer.fromarray([999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 123])))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[6, +inf)", "-5", "-4", "-3", "-2", "-1", "0", "1", "2", "3", "4", "5", "(-inf, -6]"]
        assert a.counts.counts.array.tolist() == [999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 123]

        assert a.counts[None] == 55 + 123 + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9, 10]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 10, 999]
        assert a.counts[-numpy.inf:5].tolist() == [123, 0, 1, 2, 3, 4]
        assert a.counts[-numpy.inf:numpy.inf].tolist() == [123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 999]
        assert a.counts[-numpy.inf:].tolist() == [123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        assert a.counts[5] == 5
        assert a.counts[-numpy.inf] == 123
        assert a.counts[numpy.inf] == 999
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -1]].tolist() == [7, 4, 7, 999, 5, 10]
        assert a.counts[[7, 4, 7, numpy.inf, 5, -numpy.inf, -1]].tolist() == [7, 4, 7, 999, 5, 123, 10]
        assert a.counts[[7, -numpy.inf, 4, 7, numpy.inf, 5, -numpy.inf, -1]].tolist() == [7, 123, 4, 7, 999, 5, 123, 10]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False, True]].tolist() == [0, 2, 4, 6, 8, 10]

    def test_getitem_RegularBinning(self):
        a = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(10, dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[-5, -4)", "[-4, -3)", "[-3, -2)", "[-2, -1)", "[-1, 0)", "[0, 1)", "[1, 2)", "[2, 3)", "[3, 4)", "[4, 5)"]
        assert a.counts.counts.array.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        assert a.counts[None] == sum([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9]
        assert a.counts[5] == 5
        assert a.counts[[7, 4, 7, 5, -1]].tolist() == [7, 4, 7, 5, 9]
        assert a.counts[numpy.array([7, 4, 7, 5, -1])].tolist() == [7, 4, 7, 5, 9]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False]].tolist() == [0, 2, 4, 6, 8]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False])].tolist() == [0, 2, 4, 6, 8]

        a = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5), overflow=RealOverflow(loc_overflow=RealOverflow.above1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999], dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[-5, -4)", "[-4, -3)", "[-3, -2)", "[-2, -1)", "[-1, 0)", "[0, 1)", "[1, 2)", "[2, 3)", "[3, 4)", "[4, 5)", "[5, +inf]"]
        assert a.counts.counts.array.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]

        assert a.counts[None] == sum([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 999]
        assert a.counts[5] == 5
        assert a.counts[[7, numpy.inf, 4, 7, 5, numpy.inf, -1]].tolist() == [7, 999, 4, 7, 5, 999, 9]
        assert a.counts[numpy.array([7, numpy.inf, 4, 7, 5, numpy.inf, -1])].tolist() == [7, 999, 4, 7, 5, 999, 9]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False]].tolist() == [0, 2, 4, 6, 8]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False])].tolist() == [0, 2, 4, 6, 8]

        a = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5), overflow=RealOverflow(loc_overflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[5, +inf]", "[-5, -4)", "[-4, -3)", "[-3, -2)", "[-2, -1)", "[-1, 0)", "[0, 1)", "[1, 2)", "[2, 3)", "[3, 4)", "[4, 5)"]
        assert a.counts.counts.array.tolist() == [999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        assert a.counts[None] == sum([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) + 999
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 999]
        assert a.counts[5] == 5
        assert a.counts[numpy.inf] == 999
        assert a.counts[[7, numpy.inf, 4, 7, 5, numpy.inf, -1]].tolist() == [7, 999, 4, 7, 5, 999, 9]
        assert a.counts[numpy.array([7, numpy.inf, 4, 7, 5, numpy.inf, -1])].tolist() == [7, 999, 4, 7, 5, 999, 9]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False]].tolist() == [0, 2, 4, 6, 8]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False])].tolist() == [0, 2, 4, 6, 8]

        a = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5), overflow=RealOverflow(loc_overflow=RealOverflow.below2, loc_nanflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([999, 123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["[5, +inf]", "{nan}", "[-5, -4)", "[-4, -3)", "[-3, -2)", "[-2, -1)", "[-1, 0)", "[0, 1)", "[1, 2)", "[2, 3)", "[3, 4)", "[4, 5)"]
        assert a.counts.counts.array.tolist() == [999, 123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        assert a.counts[None] == sum([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) + 999 + 123
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 999]
        assert a.counts[5] == 5
        assert a.counts[numpy.inf] == 999
        assert a.counts[numpy.nan] == 123
        assert a.counts[[7, numpy.inf, 4, 7, 5, numpy.nan, -1]].tolist() == [7, 999, 4, 7, 5, 123, 9]
        if sys.version_info[0] >= 3:
            exec("assert a.counts[[numpy.inf, ..., numpy.nan]].tolist() == [999, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 123]")
        assert a.counts[numpy.array([7, numpy.inf, 4, 7, 5, numpy.nan, -1])].tolist() == [7, 999, 4, 7, 5, 123, 9]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False]].tolist() == [0, 2, 4, 6, 8]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False])].tolist() == [0, 2, 4, 6, 8]

        a = Histogram([Axis(RegularBinning(10, RealInterval(-5, 5), overflow=RealOverflow(loc_overflow=RealOverflow.above1, loc_nanflow=RealOverflow.below1)))], UnweightedCounts(InterpretedInlineBuffer.fromarray(numpy.array([123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999], dtype=int))))
        assert a.axis[0].binning.toCategoryBinning().categories == ["{nan}", "[-5, -4)", "[-4, -3)", "[-3, -2)", "[-2, -1)", "[-1, 0)", "[0, 1)", "[1, 2)", "[2, 3)", "[3, 4)", "[4, 5)", "[5, +inf]"]
        assert a.counts.counts.array.tolist() == [123, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]

        assert a.counts[None] == sum([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) + 999 + 123
        assert a.counts[:].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert a.counts[:numpy.inf].tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 999]
        assert a.counts[5:].tolist() == [5, 6, 7, 8, 9]
        assert a.counts[5:numpy.inf].tolist() == [5, 6, 7, 8, 9, 999]
        assert a.counts[5] == 5
        assert a.counts[numpy.inf] == 999
        assert a.counts[numpy.nan] == 123
        assert a.counts[[7, numpy.inf, 4, 7, 5, numpy.nan, -1]].tolist() == [7, 999, 4, 7, 5, 123, 9]
        assert a.counts[numpy.array([7, numpy.inf, 4, 7, 5, numpy.nan, -1])].tolist() == [7, 999, 4, 7, 5, 123, 9]
        assert a.counts[[True, False, True, False, True, False, True, False, True, False]].tolist() == [0, 2, 4, 6, 8]
        assert a.counts[numpy.array([True, False, True, False, True, False, True, False, True, False])].tolist() == [0, 2, 4, 6, 8]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import functools
import operator
import re

import numpy
try:
    import pandas
except ImportError:
    raise ImportError("Install pandas package with:\n    pip install pandas\nor\n    conda install pandas")

from aghast import *

def binning2index(binning):
    if binning is None:
        return pandas.CategoricalIndex(["all"])

    elif isinstance(binning, IntegerBinning):
        if binning.loc_underflow == BinLocation.nonexistent and binning.loc_overflow == BinLocation.nonexistent:
            return pandas.RangeIndex(binning.min, binning.max + 1)
        else:
            return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, RegularBinning):
        if binning.overflow is None or (binning.overflow.loc_underflow == BinLocation.nonexistent and binning.overflow.loc_overflow == BinLocation.nonexistent and binning.overflow.loc_nanflow == BinLocation.nonexistent):
            return pandas.interval_range(binning.interval.low, binning.interval.high, binning.num, closed=("left" if binning.interval.low_inclusive else "right"))
        elif binning.overflow is None or binning.overflow.loc_nanflow == BinLocation.nonexistent:
            return binning2index(binning.toEdgesBinning())
        else:
            return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, HexagonalBinning):
        raise NotImplementedError

    elif isinstance(binning, EdgesBinning):
        if binning.overflow is None or binning.overflow.loc_nanflow == BinLocation.nonexistent:
            if binning.overflow is None or (binning.overflow.loc_underflow == BinLocation.nonexistent and binning.overflow.loc_overflow == BinLocation.nonexistent):
                return pandas.IntervalIndex.from_breaks(binning.edges, closed=("left" if binning.low_inclusive else "right"))
            elif binning.overflow is not None and binning.overflow.loc_underflow.value <= BinLocation.nonexistent.value and binning.overflow.loc_overflow.value >= BinLocation.nonexistent.value:
                edges = numpy.empty(binning._binshape()[0] + 1, dtype=numpy.float64)
                shift = int(binning.overflow.loc_underflow != BinLocation.nonexistent)
                edges[shift : shift + len(binning.edges)] = binning.edges
                if binning.overflow.loc_underflow != BinLocation.nonexistent:
                    edges[0] = -numpy.inf
                if binning.overflow.loc_overflow != BinLocation.nonexistent:
                    edges[-1] = numpy.inf
                return pandas.IntervalIndex.from_breaks(edges, closed=("left" if binning.low_inclusive else "right"))
            else:
                return binning2index(binning.toIrregularBinning())
        else:
            return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, IrregularBinning):
        if (binning.overflow is None or binning.overflow.loc_nanflow == BinLocation.nonexistent) and len(binning.intervals) != 0 and binning.intervals[0].low_inclusive != binning.intervals[0].high_inclusive and all(x.low_inclusive == binning.intervals[0].low_inclusive and x.high_inclusive == binning.intervals[0].high_inclusive for x in binning.intervals):
            left = numpy.empty(binning._binshape(), dtype=numpy.float64)
            right = numpy.empty(binning._binshape(), dtype=numpy.float64)
            flows = [] if binning.overflow is None else [(binning.overflow.loc_underflow, -numpy.inf), (binning.overflow.loc_overflow, numpy.inf)]
            low = numpy.inf
            high = -numpy.inf
            for interval in binning.intervals:
                if interval.low <= low:
                    low = interval.low
                if interval.high >= high:
                    high = interval.high
            i = 0
            for loc, val in BinLocation._belows(flows):
                if val == -numpy.inf:
                    left[i], right[i] = val, low
                if val == numpy.inf:
                    left[i], right[i] = high, val
                i += 1
            for interval in binning.intervals:
                left[i] = interval.low
                right[i] = interval.high
                i += 1
            for loc, val in BinLocation._aboves(flows):
                if val == -numpy.inf:
                    left[i], right[i] = val, low
                if val == numpy.inf:
                    left[i], right[i] = high, val
                i += 1
            return pandas.IntervalIndex.from_arrays(left, right, closed=("left" if binning.intervals[0].low_inclusive else "right"))
        else:
            return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, CategoryBinning):
        categories = []
        flows = [(binning.loc_overflow,)]
        for loc, in BinLocation._belows(flows):
            categories.append("(other)")
        categories.extend(binning.categories)
        for loc, in BinLocation._aboves(flows):
            categories.append("(other)")
        return pandas.CategoricalIndex(categories)
        
    elif isinstance(binning, SparseRegularBinning):
        return binning2index(binning.toIrregularBinning())

    elif isinstance(binning, FractionBinning):
        return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, PredicateBinning):
        return binning2index(binning.toCategoryBinning())

    elif isinstance(binning, VariationBinning):
        return binning2index(binning.toCategoryBinning())

    else:
        raise AssertionError(type(binning))

def to_pandas(obj):
    if isinstance(obj, Histogram):
        indexes = []
        names = []

        for axis in obj.allaxis:
            index = binning2index(axis.binning)

            target = None
            for i in range(len(indexes)):
                if target is None:
                    target = len(indexes[i])
                else:
                    assert target == len(indexes[i])
                indexes[i] = indexes[i].repeat(len(index))

            target = 1 if target is None else target

            # why doesn't pandas have an Index.tile function?
            x = index
            tiles = 1
            for j in range(int(numpy.floor(numpy.log2(target)))):
                x = x.append(x)
                tiles *= 2
            while tiles < target:
                x = x.append(index)
                tiles += 1

            assert len(x) == target * len(index)
            indexes.append(x)

            if axis.title is not None:
                names.append(axis.title)
            elif axis.expression is not None:
                names.append(axis.expression)
            else:
                names.append(None)

        assert len(indexes) == len(names)
        assert len(indexes) != 0
        if len(indexes) == 1:
            index = indexes[0]
        else:
            index = pandas.MultiIndex.from_arrays(indexes, names=names)

        if isinstance(obj.counts, UnweightedCounts):
            data = [obj.counts.flatarray]
            columns = ["unweighted"]
        elif isinstance(obj.counts, WeightedCounts):
            data = [obj.counts.sumw.flatarray]
            columns = ["sumw"]
            if obj.counts.sumw2 is not None:
                data.append(obj.counts.sumw2.flatarray)
                columns.append("sumw2")
            if obj.counts.unweighted is not None:
                data.append(obj.counts.unweighted.flatarray)
                columns.append("unweighted")

        if len(obj.profile) != 0:
            for i in range(len(columns)):
                columns[i] = ("counts", columns[i])

        for profile in obj.profile:
            for moment in profile.statistics.moments:
                data.append(moment.sumwxn.flatarray)
                columns.append((profile.expression, "sum" + ("w" + (repr(moment.weightpower) if moment.weightpower != 1 else "") if moment.weightpower != 0 else "") + ("x" + (repr(moment.n) if moment.n != 1 else "") if moment.n != 0 else "")))

            for quantile in profile.statistics.quantiles:
                data.append(quantile.values.flatarray)
                columns.append((profile.expression, ("p=" + ("%g" % quantile.p)) + (" (w" + (repr(quantile.weightpower) if quantile.weightpower != 1 else "") + ")" if quantile.weightpower != 0 else "")))

            if profile.statistics.mode is not None:
                data.append(profile.statistics.mode.values.flatarray)
                columns.append((profile.expression, "mode"))

            if profile.statistics.min is not None:
                data.append(profile.statistics.mode.values.flatarray)
                columns.append((profile.expression, "min"))

            if profile.statistics.max is not None:
                data.append(profile.statistics.mode.values.flatarray)
                columns.append((profile.expression, "max"))

        data = dict(zip(columns, data))

        if isinstance(columns[0], tuple):
            columns = pandas.MultiIndex.from_tuples(columns)

        return pandas.DataFrame(index=index, columns=columns, data=data)

    elif isinstance(obj, Ntuple):
        raise NotImplementedError

    elif isinstance(obj, Collection):
        raise NotImplementedError

    elif isinstance(obj, BinnedEvaluatedFunction):
        raise NotImplementedError

    else:
        raise TypeError("{0} has no pandas equivalent".format(type(obj).__name__))

def column2statistic(array, statexpr):
    m = column2statistic.moment.match(statexpr)
    if m is not None:
        if m.group(1) is None:
            weightpower = 0
        else:
            weightpower = m.group(2)
            if weightpower is None:
                weightpower = 1
            else:
                weightpower = int(weightpower)
        if m.group(3) is None:
            n = 0
        else:
            n = m.group(4)
            if n is None:
                n = 1
            else:
                n = int(n)
        return Moments(InterpretedInlineBuffer.fromarray(array), n=n, weightpower=weightpower)

    m = column2statistic.quantile.match(statexpr)
    if m is not None:
        p = float(m.group(1))
        weightpower = m.group(6)
        if weightpower is None:
            weightpower = 0
        elif weightpower == "":
            weightpower = 1
        else:
            weightpower = int(weightpower)
        return Quantiles(InterpretedInlineBuffer.fromarray(array), p=p, weightpower=weightpower)

    if statexpr == "mode":
        return Modes(InterpretedInlineBuffer.fromarray(array))

    if statexpr == "min":
        return Extremes(InterpretedInlineBuffer.fromarray(array))

    if statexpr == "max":
        return Extremes(InterpretedInlineBuffer.fromarray(array))

    return None

column2statistic.moment = re.compile(r"sum(w([+-]?\d+)?)?(x([+-]?\d+)?)?")
column2statistic.quantile = re.compile(r"p=([+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)( \(w(\d*)\))?")

def index2binning(index):
    if isinstance(index, pandas.IntervalIndex):
        ordered = (index.left[:-1] < index.left[1:]).all() and (index.right[:-1] < index.right[1:]).all()
        abutting = (index.left[1:] == index.right[:-1]).all()
        fixedwidth = len(numpy.unique(index.right - index.left)) == 1
        if index.closed == "left":
            low_inclusive, high_inclusive = True, False
        elif index.closed == "right":
            low_inclusive, high_inclusive = False, True
        elif index.closed == "both":
            low_inclusive, high_inclusive = True, True
        elif index.closed == "neither":
            low_inclusive, high_inclusive = False, False

        if ordered and abutting and fixedwidth and low_inclusive is not high_inclusive:
            return RegularBinning(len(index), RealInterval(index.left[0], index.right[-1], low_inclusive=low_inclusive, high_inclusive=high_inclusive))

        elif ordered and abutting and not fixedwidth and low_inclusive is not high_inclusive:
            edges = numpy.empty(len(index) + 1, dtype=numpy.int64)
            edges[:len(index)] = index.left
            edges[-1] = leve.right[-1]
            return EdgesBinning(edges, low_inclusive=low_inclusive, high_inclusive=high_inclusive)

        elif not abutting and fixedwidth and low_inclusive is not high_inclusive:
            bin_width = index.right[0] - index.left[0]
            origin = index.left.min()
            bins = numpy.trunc((index.left - origin) / bin_width)
            return SparseRegularBinning(bins, bin_width, origin=origin, low_inclusive=low_inclusive, high_inclusive=high_inclusive)

        else:
            return IrregularBinning([RealInterval(x.left, x.right, low_inclusive=low_inclusive, high_inclusive=high_inclusive) for x in index])

    elif isinstance(index, (pandas.Int64Index, pandas.UInt64Index)) and len(index) == 1 + index.max() - index.min():
        return IntegerBinning(index.min(), index.max())

    else:
        return CategoryBinning([str(x) for x in index])

def from_pandas(obj):
    if ((isinstance(obj.columns, pandas.MultiIndex) and "counts" in obj.columns) and ("unweighted" in obj["counts"].columns or "sumw" in obj["counts"].columns or "sumw2" in obj["counts"].columns)) or ("unweighted" in obj.columns or "sumw" in obj.columns or "sumw2" in obj.columns):
        # this is a histogram; make the data dense in the Cartesian grid
        unstacked = obj.unstack()
        unstacked = unstacked.reindex(sorted(unstacked.columns), axis=1)
        if isinstance(obj.columns, pandas.MultiIndex) and len(obj.columns.levels) == 2:
            for expression, statexpr in obj.columns:
                if statexpr == "min":
                    unstacked[expression, statexpr] = unstacked[expression, statexpr].fillna(numpy.inf)
                elif statexpr == "max":
                    unstacked[expression, statexpr] = unstacked[expression, statexpr].fillna(-numpy.inf)
                elif statexpr.startswith("sum"):
                    unstacked[expression, statexpr] = unstacked[expression, statexpr].fillna(0)

        obj = unstacked.stack(dropna=False)

        index = obj.index
        if not isinstance(index, pandas.MultiIndex):
            index = pandas.MultiIndex.from_arrays([index])

        levelsize = [len(x) for x in index.levels]
        assert functools.reduce(operator.mul, levelsize, 1) == len(obj)

        axis = []
        for leveli in range(len(index.levels)):
            level = index.get_level_values(leveli)
            sizeafter = functools.reduce(operator.mul, levelsize[leveli + 1 :], 1)
            level = level[::sizeafter][:levelsize[leveli]]
            binning = index2binning(level)
            axis.append(Axis(binning=binning, expression=level.name))
        
        unweighted, sumw, sumw2 = None, None, None
        if isinstance(obj.columns, pandas.MultiIndex) and len(obj.columns.levels) == 2 and "counts" in obj.columns:
            if "unweighted" in obj["counts"].columns:
                unweighted = InterpretedInlineBuffer.fromarray(obj["counts"]["unweighted"].values)
            if "sumw" in obj["counts"].columns:
                sumw = InterpretedInlineBuffer.fromarray(obj["counts"]["sumw"].values)
            if "sumw2" in obj["counts"].columns:
                sumw2 = InterpretedInlineBuffer.fromarray(obj["counts"]["sumw2"].values)
        else:
            if "unweighted" in obj.columns:
                unweighted = InterpretedInlineBuffer.fromarray(obj["unweighted"].values)
            if "sumw" in obj.columns:
                sumw = InterpretedInlineBuffer.fromarray(obj["sumw"].values)
            if "sumw2" in obj.columns:
                sumw2 = InterpretedInlineBuffer.fromarray(obj["sumw2"].values)

        if unweighted is not None and sumw is None and sumw2 is None:
            counts = UnweightedCounts(unweighted)
        elif sumw is not None:
            counts = WeightedCounts(sumw, sumw2, None if unweighted is None else UnweightedCounts(unweighted))
        else:
            raise ValueError("unrecognized combination of counts: ".format(", ".join(s for s, x in [("unweighted", unweighted), ("sumw", sumw), ("sumw2", sumw2)] if x is not None)))

        statistics = {}
        if isinstance(obj.columns, pandas.MultiIndex) and len(obj.columns.levels) == 2:
            for expression, statexpr in obj.columns:
                if expression != "counts":
                    array = obj[expression][statexpr].values

                    statistic = statistics.get(expression, None)
                    if statistic is None:
                        statistic = statistics[expression] = Statistics()

                    data = column2statistic(array, statexpr)
                    if isinstance(data, Moments):
                        statistic.moments = list(statistic.moments) + [data]
                    elif isinstance(data, Quantiles):
                        statistic.quantiles = list(statistic.quantiles) + [data]
                    elif isinstance(data, Modes):
                        statistic.mode = data
                    elif statexpr == "min":
                        statistic.min = data
                    elif statexpr == "max":
                        statistic.max = data

        else:
            for expression in obj.columns:
                if expression != "unweighted" and expression != "sumw" and expression != "sumw2":
                    array = obj[expression].values
                    statistics[expression] = Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(array), n=1, weightpower=(0 if sumw2 is None else 1))])

        profile = []
        for expression, statistic in statistics.items():
            if len(statistic.moments) != 0 or len(statistic.quantiles) != 0 or statistic.mode is not None or statistic.min is not None or statistic.max is not None:
                profile.append(Profile(expression, statistic))

        return Histogram(axis, counts, profile=profile)

    else:
        raise TypeError("DataFrame not recognized as a histogram")



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import os
import shutil

if __name__ == "__main__":
    if os.path.exists(os.path.join("aghast", "aghast_generated")):
        shutil.rmtree(os.path.join("aghast", "aghast_generated"))

    os.chdir("aghast")
    os.system("flatc --python ../../flatbuffers/aghast.fbs")

    with open(os.path.join("aghast_generated", "StatisticFilter.py")) as f:
        tmp = f.read().replace("inf.0", "float('inf')")
    with open(os.path.join("aghast_generated", "StatisticFilter.py"), "w") as f:
        f.write(tmp)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

# ///////////////////////////////////////////////// binning
class BinLocation(object):
    loc_below3 = -3
    loc_below2 = -2
    loc_below1 = -1
    loc_nonexistent = 0
    loc_above1 = 1
    loc_above2 = 2
    loc_above3 = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class NonRealMapping(object):
    missing = 0
    in_underflow = 1
    in_overflow = 2
    in_nanflow = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

# ///////////////////////////////////////////////// decoration
class DecorationLanguage(object):
    deco_unspecified = 0
    deco_css = 1
    deco_vega = 2
    deco_json = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Moments(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMoments(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Moments()
        x.Init(buf, n + offset)
        return x

    # Moments
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Moments
    def SumwxnType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Moments
    def Sumwxn(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Moments
    def N(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Moments
    def Weightpower(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Moments
    def Filter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .StatisticFilter import StatisticFilter
            obj = StatisticFilter()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def MomentsStart(builder): builder.StartObject(5)
def MomentsAddSumwxnType(builder, sumwxnType): builder.PrependUint8Slot(0, sumwxnType, 0)
def MomentsAddSumwxn(builder, sumwxn): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(sumwxn), 0)
def MomentsAddN(builder, n): builder.PrependInt8Slot(2, n, 0)
def MomentsAddWeightpower(builder, weightpower): builder.PrependInt8Slot(3, weightpower, 0)
def MomentsAddFilter(builder, filter): builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(filter), 0)
def MomentsEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class Metadata(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMetadata(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Metadata()
        x.Init(buf, n + offset)
        return x

    # Metadata
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Metadata
    def Data(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Metadata
    def Language(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def MetadataStart(builder): builder.StartObject(2)
def MetadataAddData(builder, data): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0)
def MetadataAddLanguage(builder, language): builder.PrependUint32Slot(1, language, 0)
def MetadataEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import numpy
try:
    import fnal_column_analysis_tools.hist as hist
except ImportError:
    raise ImportError("Install fnal_column_analysis_tools package with:\n    pip install fnal_column_analysis_tools")

from aghast import *

_ovf_convention = lambda: RealOverflow(loc_underflow=BinLocation.below1,
                                       loc_overflow=BinLocation.above1,
                                       loc_nanflow=BinLocation.above2)

def from_fnalhist(obj):
    if not isinstance(obj, hist.Hist):
        raise TypeError("cannot convert {0} to an aghast histogram".format(type(obj).__name__))
    axes = []
    expanded_shape = []
    expanded_slices = []
    expansion_offsets = []
    for i, ax in enumerate(obj.axes()):
        if isinstance(ax, hist.Cat):
            cats = obj.identifiers(ax)
            expanded_shape.append(len(cats))
            expanded_slices.append(numpy.array(cats))
            expansion_offsets.append(i)
            axes.append(Axis(CategoryBinning(categories=cats), expression=ax.name, title=ax.label))
        elif isinstance(ax, hist.Bin) and ax._uniform:
            axes.append(Axis(RegularBinning(ax._bins,
                            RealInterval(ax._lo, ax._hi),
                            overflow=_ovf_convention()
                            ), expression=ax.name, title=ax.label))
            expanded_shape.append(ax.size)
        elif isinstance(ax, hist.Bin) and not ax._uniform:
            axes.append(Axis(EdgesBinning(ax._bins[:-1],
                                        overflow=_ovf_convention()
                                        ), expression=ax.name, title=ax.label))
            expanded_shape.append(ax.size)
        else:
            raise NotImplementedError("unknown fnalhist Axis type")

    expanded_sumw = numpy.zeros(shape=expanded_shape, dtype=obj._dtype)
    if obj._sumw2 is not None:
        expanded_sumw2 = numpy.zeros(shape=expanded_shape, dtype=obj._dtype)

    walk_shape = [len(s) for s in expanded_slices]
    walk_indices = numpy.unravel_index(numpy.arange(numpy.prod(walk_shape)), walk_shape)
    extract_slices = list(zip(*(v[k] for v,k in zip(expanded_slices, walk_indices))))
    for iwalk, tup in enumerate(extract_slices):
        if tup in obj._sumw:
            insert_slice = [slice(None) for _ in expanded_shape]
            for walk_dim, ins_dim in enumerate(expansion_offsets):
                insert_slice[ins_dim] = walk_indices[walk_dim][iwalk]
            insert_slice = tuple(insert_slice)
            expanded_sumw[insert_slice] = obj._sumw[tup]
            if obj._sumw2 is not None:
                expanded_sumw2[insert_slice] = obj._sumw2[tup]

    if obj._sumw2 is not None:
        counts = WeightedCounts(
            sumw=InterpretedInlineBuffer.fromarray(expanded_sumw),
            sumw2=InterpretedInlineBuffer.fromarray(expanded_sumw2)
        )
    else:
        counts = UnweightedCounts(InterpretedInlineBuffer.fromarray(expanded_sumw))

    hout = Histogram(axis=axes, title=obj.label, counts=counts)
    return hout

def to_fnalhist(obj):
    if not isinstance(obj, Histogram):
        raise TypeError("cannot convert {0} to a fnal_column_analysis_tools histogram".format(type(obj).__name__))

    axes = []
    sparse_binning = {}
    dense_map = {}
    dense_shape = []
    for i, ax in enumerate(obj.axis):
        if isinstance(ax.binning, CategoryBinning):
            sparse_binning[i] = ax.binning.categories
            if ax.binning.loc_overflow.value < BinLocation.nonexistent.value:
                sparse_binning[i].insert(0, "")
            elif ax.binning.loc_overflow.value > BinLocation.nonexistent.value:
                sparse_binning[i].append("")
            new_ax = hist.Cat(ax.expression, ax.title)
            new_ax._categories = sparse_binning[i]
            axes.append(new_ax)
        elif isinstance(ax.binning, RegularBinning):
            dense_map[i] = [-numpy.inf, ..., numpy.inf, numpy.nan]
            dense_shape.append(ax.binning.num+3)
            axes.append(hist.Bin(ax.expression,
                                 ax.title,
                                 ax.binning.num,
                                 ax.binning.interval.low,
                                 ax.binning.interval.high
                                ))
        elif isinstance(ax.binning, EdgesBinning):
            dense_map[i] = [-numpy.inf, ..., numpy.inf, numpy.nan]
            dense_shape.append(ax.binning.edges.size+2)
            axes.append(hist.Bin(ax.expression,
                                 ax.title,
                                 ax.binning.edges
                                ))
        else:
            raise TypeError("unable to convert axes of type {0} to fnalhist axes".format(type(ax).__name__))

    hout = hist.Hist(obj.title, *axes)
    if isinstance(obj.counts, WeightedCounts):
        hout._init_sumw2()

    walk_shape = [len(s) for s in sparse_binning.values()]
    walk_indices = numpy.unravel_index(numpy.arange(numpy.prod(walk_shape)), walk_shape)
    for walk_index in zip(*walk_indices):
        insert_index = tuple(v[k] for k,v in zip(walk_index, sparse_binning.values()))
        extract_index = [None]*len(axes)
        for k,v in dense_map.items():
            extract_index[k] = v
        for k,v in zip(sparse_binning.keys(), walk_index):
            extract_index[k] = v
        extract_index = tuple(extract_index)
        if isinstance(obj.counts, UnweightedCounts):
            hout._sumw[insert_index] = obj.counts[extract_index]
        elif isinstance(obj.counts, WeightedCounts):
            counts = obj.counts[extract_index]
            hout._sumw[insert_index] = counts['sumw']
            hout._sumw2[insert_index] = counts['sumw2']
    
    return hout



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class FractionErrorMethod(object):
    frac_unspecified = 0
    frac_normal = 1
    frac_clopper_pearson = 2
    frac_wilson = 3
    frac_agresti_coull = 4
    frac_feldman_cousins = 5
    frac_jeffrey = 6
    frac_bayesian_uniform = 7




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class OverlappingFillStrategy(object):
    overfill_unspecified = 0
    overfill_all = 1
    overfill_first = 2
    overfill_last = 3




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

import flatbuffers

class RawExternalBuffer(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRawExternalBuffer(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RawExternalBuffer()
        x.Init(buf, n + offset)
        return x

    # RawExternalBuffer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RawExternalBuffer
    def Pointer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # RawExternalBuffer
    def Numbytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # RawExternalBuffer
    def ExternalSource(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # RawExternalBuffer
    def Location(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def RawExternalBufferStart(builder): builder.StartObject(4)
def RawExternalBufferAddPointer(builder, pointer): builder.PrependUint64Slot(0, pointer, 0)
def RawExternalBufferAddNumbytes(builder, numbytes): builder.PrependUint64Slot(1, numbytes, 0)
def RawExternalBufferAddExternalSource(builder, externalSource): builder.PrependUint32Slot(2, externalSource, 0)
def RawExternalBufferAddLocation(builder, location): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(location), 0)
def RawExternalBufferEnd(builder): return builder.EndObject()



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/aghast/blob/master/LICENSE

import unittest

import pytest
import numpy

import aghast
from aghast import *

pandas = pytest.importorskip("pandas")
pytest.importorskip("aghast._connect._pandas")

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_pandas(self):
        a = Histogram([Axis(IntegerBinning(5, 9)), Axis(IntegerBinning(-5, 5))], WeightedCounts(InterpretedInlineBuffer.fromarray(numpy.arange(55)), sumw2=InterpretedInlineBuffer.fromarray(numpy.arange(55))), profile=[Profile("qqq", Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.arange(55)), n=2, weightpower=2), Moments(InterpretedInlineBuffer.fromarray(numpy.arange(55)), n=0)], quantiles=[Quantiles(InterpretedInlineBuffer.fromarray(numpy.arange(55)), p=0.5)], max=Extremes(InterpretedInlineBuffer.fromarray(numpy.arange(55))), min=Extremes(InterpretedInlineBuffer.fromarray(numpy.arange(55))), mode=Modes(InterpretedInlineBuffer.fromarray(numpy.arange(55))))), Profile("zzz", Statistics(moments=[Moments(InterpretedInlineBuffer.fromarray(numpy.arange(55)), n=2, weightpower=2), Moments(InterpretedInlineBuffer.fromarray(numpy.arange(55)), n=0)], quantiles=[Quantiles(InterpretedInlineBuffer.fromarray(numpy.arange(55)), p=1)], max=Extremes(InterpretedInlineBuffer.fromarray(numpy.arange(55))), min=Extremes(InterpretedInlineBuffer.fromarray(numpy.arange(55))), mode=Modes(InterpretedInlineBuffer.fromarray(numpy.arange(55)))))])
        aghast.from_pandas(aghast.to_pandas(a)[:25])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: aghast_generated

class FunctionObjectData(object):
    NONE = 0
    ParameterizedFunction = 1
    BinnedEvaluatedFunction = 2

